#coding=utf-8


class Store(object):
    def __init__(self):
        pass

    def save_seed_page(self):
        pass

    def save_list_page(self):
        pass

    def save_detail_page(self):
        pass

    def get_list_page(self):
        pass

    def get_detail_page(self):
        pass



from scrapy.crawler import CrawlerProcess
from scrapy.utils.project import get_project_settings
import os

if __name__ == "__main__":
    cwd = os.path.dirname(__file__)

    os.environ.setdefault("XDG_CONFIG_HOME", cwd + "/../")
    settings = get_project_settings()
    process = CrawlerProcess(settings)
    # 'followall' is the name of one of the spiders of the project.
    process.crawl("website",
                  config=cwd + "/config/re_example.yaml",
                  # config=cwd + "/config/allsharktankproducts.yaml",
                  # url="https://allsharktankproducts.com/sharktankproducts/season-12-products/",
                  # LOG_LEVEL="DEBUG",
                  # DOWNLOAD_DELAY=3
                  COOKIES_ENABLED=False
                  )
    process.start()
    process.stop()



# coding=utf-8
import logging
import jsonpath
import re

from scrapy.selector import SelectorList
from dig_spider.engine.config import DigConfig


logger = logging.getLogger(__name__)


class ExtractEngine(object):
    def __init__(self, config_file):
        self.config = DigConfig(config_file)

    def extract_page(self, response):
        page_config = self.config.get_page_config(response.url)
        if page_config is None:
            logger.error("not find page extract config for url: %s", response.url)
            return [], ''
        list_items = self.extract_by_path(page_config.list_rule, response) if page_config.list_rule else []

        results = []
        list_items = list_items if list_items else [response]
        for item in list_items:
            result = self.extract_item(page_config.item_rules, item)
            if page_config.page_type:
                result['page_type'] = page_config.page_type
            results.append(result)

        next_page_url = self.extract_by_path(page_config.next_page_rule, response) if page_config.next_page_rule else ''
        if isinstance(next_page_url, SelectorList):
            next_page_url = next_page_url.get()
        next_page_url = response.urljoin(next_page_url) if next_page_url else ''
        if page_config.code:
            exec(page_config.code)
        return results, next_page_url

    def extract_by_path(self, rule, response):
        if rule.path_type == 'css' and rule.path:
            return response.css(rule.path)
        elif rule.path_type == 'xpath' and rule.path:
            return response.xpath(rule.path)
        elif rule.path_type == 're' and rule.path:
            return eval(rule.path)
        elif rule.path_type == 'json' and rule.path:
            if type(response) != dict:
                response = response.json()
            return jsonpath.jsonpath(response, rule.path)[0]
        elif rule.path_type == 'code' and rule.path:
            return eval(rule.path)
        logger.error("name: %s, path: %s not work for url: %s", rule.name, rule.path, response.url)
        return None

    def extract_item(self, rules, response):
        item = {}
        for key, rule in rules.items():
            node = self.extract_by_path(rule, response)
            # node = response if node is None else node
            value = node.get() if isinstance(node, SelectorList) else node
            if not value:
                logger.info("%s: %s  extract empty", key, rule.path)
            item[key] = rule.process_funcs(value) if rule.funcs else value
        return item



# Scrapy settings for dig_spider project
#
# For simplicity, this file contains only settings considered important or
# commonly used. You can find more settings consulting the documentation:
#
#     https://docs.scrapy.org/en/latest/topics/settings.html
#     https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
#     https://docs.scrapy.org/en/latest/topics/spider-middleware.html

BOT_NAME = "dig_spider"

LOG_LEVEL = 'INFO'
SPIDER_MODULES = ["dig_spider.spiders"]
NEWSPIDER_MODULE = "dig_spider.spiders"


# Crawl responsibly by identifying yourself (and your website) on the user-agent
USER_AGENT = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/129.0.0.0 Safari/537.36"


# Obey robots.txt rules
ROBOTSTXT_OBEY = False

# Configure maximum concurrent requests performed by Scrapy (default: 16)
#CONCURRENT_REQUESTS = 32

# Configure a delay for requests for the same website (default: 0)
# See https://docs.scrapy.org/en/latest/topics/settings.html#download-delay
# See also autothrottle settings and docs
# DOWNLOAD_DELAY = 3
# The download delay setting will honor only one of:
#CONCURRENT_REQUESTS_PER_DOMAIN = 16
#CONCURRENT_REQUESTS_PER_IP = 16

# Disable cookies (enabled by default)
# COOKIES_ENABLED = False

# Disable Telnet Console (enabled by default)
#TELNETCONSOLE_ENABLED = False

# Override the default request headers:
#DEFAULT_REQUEST_HEADERS = {
#    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
#    "Accept-Language": "en",
#}

# Enable or disable spider middlewares
# See https://docs.scrapy.org/en/latest/topics/spider-middleware.html
SPIDER_MIDDLEWARES = {
   "dig_spider.middlewares.DigEngineSpiderMiddleware": 543,
}

# Enable or disable downloader middlewares
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html
DOWNLOADER_MIDDLEWARES = {
   "dig_spider.middlewares.DigEngineDownloaderMiddleware": 543,
}

# Enable or disable extensions
# See https://docs.scrapy.org/en/latest/topics/extensions.html
#EXTENSIONS = {
#    "scrapy.extensions.telnet.TelnetConsole": None,
#}

# Configure item pipelines
# See https://docs.scrapy.org/en/latest/topics/item-pipeline.html
ITEM_PIPELINES = {
   "dig_spider.pipelines.DigEnginePipeline": 300,
}

# Enable and configure the AutoThrottle extension (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/autothrottle.html
#AUTOTHROTTLE_ENABLED = True
# The initial download delay
#AUTOTHROTTLE_START_DELAY = 5
# The maximum download delay to be set in case of high latencies
#AUTOTHROTTLE_MAX_DELAY = 60
# The average number of requests Scrapy should be sending in parallel to
# each remote server
#AUTOTHROTTLE_TARGET_CONCURRENCY = 1.0
# Enable showing throttling stats for every response received:
#AUTOTHROTTLE_DEBUG = False

# Enable and configure HTTP caching (disabled by default)
# See https://docs.scrapy.org/en/latest/topics/downloader-middleware.html#httpcache-middleware-settings
#HTTPCACHE_ENABLED = True
#HTTPCACHE_EXPIRATION_SECS = 0
#HTTPCACHE_DIR = "httpcache"
#HTTPCACHE_IGNORE_HTTP_CODES = []
#HTTPCACHE_STORAGE = "scrapy.extensions.httpcache.FilesystemCacheStorage"

# Set settings whose default value is deprecated to a future-proof value
REQUEST_FINGERPRINTER_IMPLEMENTATION = "2.7"
TWISTED_REACTOR = "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
FEED_EXPORT_ENCODING = "utf-8"



import logging
from typing import Any

import scrapy

from scrapy.crawler import Crawler
from scrapy.http import Response
from dig_spider.engine.extract import ExtractEngine


class WebsiteSpider(scrapy.Spider):
    name = "website"
    custom_settings = {}

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        config_file = getattr(self, 'config', "")
        if not config_file:
            raise Exception("No config, please set -a config=xxx")
        self.extract = ExtractEngine(config_file)
        self.config = self.extract.config
        url = getattr(self, 'url', "")
        if url:
            self.start_urls = [url]
        else:
            self.start_urls = self.config.get_start_urls()
        self.allowed_domains = self.config.get_allowed_domains()

    def parse(self, response: Response, **kwargs: Any) -> Any:
        items, next_page_url = self.extract.extract_page(response)
        if next_page_url:
            self.logger.info("next_page: %s", next_page_url)
            yield response.follow(next_page_url, callback=self.parse)

        for item in items:
            if 'url' in item:
                item['url'] = response.urljoin(item['url'])
                self.logger.info("item page: %s", item['url'])
                yield scrapy.Request(item['url'], callback=self.parse)
            yield item

    @classmethod
    def from_crawler(cls, crawler: Crawler, *args: Any, **kwargs: Any):
        spider = super().from_crawler(crawler,  *args, **kwargs)
        if kwargs:
            cls.custom_settings.update(kwargs)
            cls.update_settings(spider.settings)
            logging.info("config custom settings: %s", cls.custom_settings)
        return spider



# coding=utf-8
import logging
import re
import yaml
import os

logger = logging.getLogger(__name__)


class DigConfig(object):
    def __init__(self, config_file):
        self.config_file = config_file
        self.config = self.__get_yaml_data()
        self.__valid_config()
        self.__parse_config()

    def __get_yaml_data(self):
        with open(self.config_file, encoding='utf-8') as file:
            content = file.read()
            data = yaml.load(content, Loader=yaml.FullLoader)
            return data

    def __valid_config(self):
        if self.config is None:
            raise Exception("config is empty")

        if 'start_urls' not in self.config:
            raise Exception("start_urls is not in config")

        if 'extract' not in self.config:
            raise Exception("extract is not in config")

        for item in self.config.get('extract'):
            if "page" in item:
                for key in ["url_patterns", "item_rules"]:
                    if key not in item.get("page"):
                        raise Exception("{} is must in page config".format(key))
            else:
                raise Exception("page is must in extract config")

    def __parse_config(self):
        self.pages = []
        cwd = os.path.dirname(self.config_file)
        for page_config in self.config.get('extract'):
            self.pages.append(Page(page_config.get('page'), cwd))

    def get_start_urls(self):
        return self.config.get('start_urls')

    def get_allowed_domains(self):
        return self.config.get('allowed_domains', [])

    def get_proxies(self):
        return self.config.get('proxies', [])

    def get_page_config(self, url):
        for page in self.pages:
            for url_pattern in page.url_patterns:
                if re.match(url_pattern, url):
                    return page
        return None

    def get_headers(self):
        return self.config.get("headers", None)


class Page(object):
    def __init__(self, config, cwd):
        self.config = config
        self.cwd = cwd
        self.__parse_config()

    def __parse_config(self):
        self.page_type = self.config.get('page_type', '')
        code = self.config.get('code', '')
        if code:
            lines = code.split('\n')
            code = ''
            for line in lines:
                line = line.strip(' ') + '\n'
                code += line
        self.code = code
        if 'code_file' in self.config:
            with open(self.cwd + '/' + self.config.get('code_file'), encoding='utf-8') as file:
                self.code = file.read() + '\n' + self.code

        self.url_patterns = [re.compile(url_pattern) for url_pattern in self.config.get('url_patterns')]
        self.next_page_rule = Rule('next_page_rule',
                                   self.config.get('next_page_rule')) if 'next_page_rule' in self.config else None
        self.list_rule = Rule('list_rule', self.config.get('list_rule')) if 'list_rule' in self.config else None
        self.item_rules = {key: Rule(key, value) for key, value in self.config.get('item_rules', {}).items()}


class Rule(object):
    def __init__(self, name, rule):
        self.name = name
        if rule:
            self.__parse_path(rule)
        else:
            raise Exception("path is empty")

    def __parse_path(self, rule):
        if '{' not in rule:
            raise Exception("rule path is invalid")
        self.path_type, rule = rule.split('{', 1)
        self.path, self.funcs = rule.split('}')

    def get_path_funcs(self, rule, path_type='css'):
        rule = rule.replace(path_type + '{', '')
        path, funcs = rule.split('}')
        return path, funcs

    def process_funcs(self, value):

        try:
            digstr = DigString(value)
            digstr = eval('digstr' + self.funcs)
            return digstr.value
        except Exception as e:
            logging.error("%s is not work for: %s", self.funcs, value)
            return value


class DigString(object):
    def __init__(self, value):
        self.value = value

    def re(self, rule, index=0):
        groups = re.findall(rule, self.value)
        if groups:
            return DigString(groups[index])
        return self

    def replace(self, old, new, count=-1):
        return DigString(self.value.replace(old, new, count))

    def split(self, sep=None, maxsplit=-1):
        return [DigString(v) for v in self.value.split(sep, maxsplit)]

    def clear_html(self):
        '''
         clear html tag,
        :return:
        '''
        pass
